//------------------------------------------------------------------------
//
// OpenDCC - DCC Sniffer
//
// Copyright (c) 2008 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      led_control.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-03-02 V0.1 kw started
//
//------------------------------------------------------------------------
//
// purpose:   DCC sniffer
//            here: timing and status engine for LEDs
//
// content:   1. Interface
//               calls for on, off, blinking, flashing, ...
//            2  runtime-part
//               must be called every time_tick, needs systick_1ms
//
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "config.h"
#include "hardware.h"
#include "timerPWM.h"

#define SIMULATION  0            // 0: real application
                                 // 1: test receive routine
                                 // 2: test timing engine
                                 // 3: test action

//=============================================================================
// 1. Definitions
//=============================================================================
//
//
// #define LED_PROG     0     // program   - yellow
// #define LED_RAILCOM  1     // railcom   - yellow
// #define LED_ACC      2     // accessory - green
// #define LED_LOCO     3     // loco      - green
// #define LED_GO       4     // sniffing  - blue
// #define LED_ON       5     // on        - green


#define OFF     0
#define ON      1

#define LED_INPORT      PINA
#define LED_OUTPORT     PORTA

volatile struct
  {
    unsigned char rest;         // time until next change in Ticks (20ms)
    unsigned char ontime;       // Einschaltzeit
    unsigned char offtime;      // Ausschaltzeit
    unsigned char pause;    
    unsigned char flashes;     // Anzahl Pulse
    unsigned char act_flash;    
  } led[6];

signed char led_tick;           // to set up timing

#define LED_TICK  5000L       // given in us

//------------------------------------------------------------------
void output(unsigned char port_no, unsigned char state)
  {
    if (state == 0)
      {                              
        LED_OUTPORT &= (~(1<<port_no));
      }
    else
      {
        LED_OUTPORT |= (1<<port_no);
      }
  }


void led_on(unsigned char myled)
  {
    led[myled].rest = 0;
    output(myled,ON);
  }

void led_half(unsigned char myled)
  {
    led[myled].act_flash = 1;
    led[myled].flashes = 1;
    led[myled].pause   = 3;
    led[myled].offtime = 3;  // 25% duty 
    led[myled].ontime  = 1;  
    led[myled].rest    = 1;
    output(myled,ON);
  }

// make a single fast pulse, then dark
void led_fast_pulse(unsigned char myled)
  {
    led[myled].act_flash = 0;
    led[myled].flashes = 0;
    led[myled].pause   =      0L / LED_TICK;
    led[myled].offtime =      0L / LED_TICK;
    led[myled].ontime  =      0L / LED_TICK;
    led[myled].rest    = 100000L / LED_TICK;
    output(myled,ON);
  }

// make a single fast pulse (and return to previous state)
void led_highlight(unsigned char myled)
  {
    if (led[myled].rest)  // if anything is active - add an extended active state
      {
        led[myled].rest   =  100000L / LED_TICK;;
        output(myled,ON);
      }
    else led_fast_pulse(myled);
  }


void led_off(unsigned char myled)
  {
    led[myled].rest   =  0;
    output(myled,OFF);
  }


// make a single pulse
void led_pulse(unsigned char myled)
  {
    led[myled].act_flash = 0;
    led[myled].flashes = 0;
    led[myled].pause   =      0L / LED_TICK;
    led[myled].offtime =      0L / LED_TICK;
    led[myled].ontime  =      0L / LED_TICK;
    led[myled].rest    = 500000L / LED_TICK;
    output(myled,ON);
  }

// make a single pulse, dimmed
void led_half_pulse(unsigned char myled)
  {
    if (led[myled].rest) return;   // prevent reentry
    
    led[myled].act_flash = 1;      // ev. vorziehen
    led[myled].flashes = 120000L / 4 / LED_TICK;   // use the flasher to determine time;
    led[myled].pause   = 0;
    led[myled].offtime = 3; // 25% duty cycle
    led[myled].ontime  = 1;
    led[myled].rest    = 1;
    output(myled,ON);
  }



// make a double pulse
void led_double_pulse(unsigned char myled)
  {
    led[myled].act_flash = 1;
    led[myled].flashes = 2;
    led[myled].pause   =      0L / LED_TICK;
    led[myled].offtime = 240000L / LED_TICK;
    led[myled].ontime  = 100000L / LED_TICK;
    led[myled].rest    = 100000L / LED_TICK;
    output(myled,ON);
  }

// slow blinking
void led_blinking(unsigned char myled)
  {
    led[myled].act_flash = 1;
    led[myled].flashes = 1;
    led[myled].pause   = 600000L / LED_TICK;
    led[myled].offtime = 600000L / LED_TICK;
    led[myled].ontime  = 400000L / LED_TICK;
    led[myled].rest    = 400000L / LED_TICK;
    output(myled,ON);
  }

void led_XoooXooo(unsigned char myled)
  {
    led[myled].act_flash = 1;
    led[myled].flashes = 1;
    led[myled].pause   = 500000L / LED_TICK;
    led[myled].offtime = 500000L / LED_TICK;
    led[myled].ontime  = 200000L / LED_TICK;
    led[myled].rest    = 200000L / LED_TICK;
    output(myled,ON);
  }

void led_oXXXoXXX(unsigned char myled)
  {
    led[myled].act_flash = 1;
    led[myled].flashes = 1;
    led[myled].pause   = 200000L / LED_TICK;
    led[myled].offtime = 200000L / LED_TICK;
    led[myled].ontime  = 500000L / LED_TICK;
    led[myled].rest    = 500000L / LED_TICK;
    output(myled,ON);
  }
void led_XoooooXooooo(unsigned char myled)
  {
    led[myled].act_flash = 1;
    led[myled].flashes = 1;
    led[myled].pause   = 650000L / LED_TICK;
    led[myled].offtime = 650000L / LED_TICK;
    led[myled].ontime  =  50000L / LED_TICK;
    led[myled].rest    =  50000L / LED_TICK;
    output(myled,ON);
  }

void led_oXXXXXoXXXXX(unsigned char myled)
  {
    led[myled].act_flash = 1;
    led[myled].flashes = 1;
    led[myled].pause   = 100000L / LED_TICK;
    led[myled].offtime = 100000L / LED_TICK;
    led[myled].ontime  = 600000L / LED_TICK;
    led[myled].rest    = 600000L / LED_TICK;
    output(myled,ON);
  }



// make a series of flashes, then a longer pause
void led_fast_flash(unsigned char myled, unsigned char count)
  {
    led[myled].act_flash = 1;
    led[myled].flashes = count;
    led[myled].pause   = 700000L / LED_TICK;
    led[myled].offtime = 240000L / LED_TICK;
    led[myled].ontime  = 120000L / LED_TICK;
    led[myled].rest    = 120000L / LED_TICK;
    output(myled,ON);
  }

void init_led(void)
  {
    led_blinking(LED_ON);
    led_double_pulse(LED_PROG);
    led_double_pulse(LED_RAILCOM);
    led_double_pulse(LED_GO);
    led_double_pulse(LED_ACC); 
    led_double_pulse(LED_LOCO);
    led_tick = 0;
  }

void run_led(void)             // to be called in main loop every tick
  {
    unsigned char i;

    signed char compare;       // keep this at 8 bit to force wrap around

    compare = led_tick - systick_1ms;
    if (compare >= (LED_TICK/1000L))
      {
        led_tick += (LED_TICK/1000L);

        for (i=0; i<6; i++)
          {                         // now process new state for this led

            if (led[i].rest != 0)
              {
                if (--led[i].rest == 0)
                  {
                     if (LED_INPORT & (1<<i))
                      {
                        if (led[i].act_flash == led[i].flashes)
                          {
                            led[i].rest = led[i].pause;
                            led[i].act_flash = 0;
                          }
                        else
                          {
                            led[i].rest = led[i].offtime;
                          }
                        output(i,OFF);
                      }
                    else
                      {
                        led[i].act_flash++;
                        led[i].rest = led[i].ontime;
                        output(i,ON);
                      }
                  }
              }
          }
      }
  }
