//----------------------------------------------------------------
//
// OpenDCC
//
// Copyright (c) 2006-2008 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      rs485.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2008-07-08 V0.1 copied from DCCsniffer, modified for
//                            Xpressnet
//            2008-07-24 V0.2 changed rx to int
//
//-----------------------------------------------------------------
//
// purpose:   lowcost central station for dcc
// content:   routines for RS485
//            see also:
//            xpressnet definition
//            here: RX and TX Fifos
//                  Control of line direction 
//
//-----------------------------------------------------------------


#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>         // put var to program memory
#include <avr/eeprom.h>
#include <util/parity.h> 
#include <string.h>

#include "config.h"               // general structures and definitions
#include "hardware.h"
#include "rs485.h"

//
//=====================================================================
//
// RS485
//
// purpose:   send and receive messages from xpressnet
//
// how:       uart acts with interrupt on fifos.
//            ohter programs access only the fifos.
//            hardware flowcontrol (direction pin)
//
// interface: see rs485.h
//
//-----------------------------------------------------------------


// FIFO objects for RX and TX
// max. Size: 255

#define X_RxBuffer_Size  64              // mind. 16
unsigned int X_RxBuffer[X_RxBuffer_Size];


#define X_TxBuffer_Size  64
unsigned char X_TxBuffer[X_TxBuffer_Size];

unsigned char X_rx_read_ptr = 0;        // point to next read
unsigned char X_rx_write_ptr = 0;       // point to next write
unsigned char X_rx_fill = 0;
unsigned char X_tx_read_ptr = 0;
unsigned char X_tx_write_ptr = 0;
unsigned char X_tx_fill = 0;


//-------------------------------------------------------------------
// this goes to usart 1
//-------------------------------------------------------------------

// support

void XP_flush_rx(void)                 // Flush Receive-Buffer
  {
    uint8_t sreg = SREG;
    cli();
    do
      {
        UDR1;
      }
    while (UCSR1A & (1 << RXC1));

    // Rcksetzen von Receive Complete-Flags
    UCSR1A |= (1 << RXC1);
    X_rx_read_ptr = 0;      
    X_rx_write_ptr = 0;
    X_rx_fill = 0; 
    SREG = sreg;     
  }

void set_XP_to_receive(void)
  {
    // for Xpressnet Adapter: Set MOSI to LOW
    // PORTB &= ~(1 << PB5);
    // XP_flush_rx();
  }

void set_XP_to_transmit(void)
  {
    // for Xpressnet Adapter: Set MOSI to HIGH
    // PORTB |= (1 << PB5);
  }

void X_init_rs485(void)
  {
    uint16_t ubrr;
    uint8_t sreg = SREG;

    // DDRB |= (1 << PB5);          // direction control, set as output

    cli();

    UCSR1B = 0;                  // stop everything

    // we calculate 100* and add 50 to get correct integer cast;
    // normal speed - pre divider 16 (high speed mode - pre divider 8)

    // ubrr = (uint16_t) ((uint32_t) F_CPU/(16*62500L) - 1);    
    ubrr = (uint16_t) ((uint32_t) (F_CPU/(16*625L) - 100L + 50L) / 100);
    UBRR1H = (uint8_t) (ubrr>>8);
    UBRR1L = (uint8_t) (ubrr);
    UCSR1A = (1 << RXC1) | (1 << TXC1) | (0 << U2X1);  // U2X1: Speed Mode
   
    // FIFOs fr Ein- und Ausgabe initialisieren

    X_tx_read_ptr = 0;
    X_tx_write_ptr = 0;
    X_tx_fill = 0;

    // UART Receiver und Transmitter anschalten, Receive-Interrupt aktivieren
    // Data mode 9N1, asynchron
    
    UCSR1B = (1 << RXCIE1)
           | (0 << TXCIE1)
           | (0 << UDRIE1)
           | (1 << RXEN1) 
           | (1 << TXEN1) 
           | (1 << UCSZ12)         // set frame length to 9bit;
           | (0 << RXB81)	
           | (1 << TXB81);         // set 9th bit to 1


    UCSR1C = (1 << URSEL1)        // must be 1 (selector)
           | (0 << UMSEL1)        // 0 = asyn mode
           | (0 << UPM11)         // 00 = parity disabled
           | (0 << UPM10)         // 
           | (0 << USBS1)         // 1 = tx with 2 stop bits
           | (1 << UCSZ11)        // 11 = 8 or 9 bits
           | (1 << UCSZ10)
           | (0 << UCPOL1);

    XP_flush_rx();
    
    UCSR1A |= (1 << TXC1);         // clear tx
    
    SREG = sreg;
}


//---------------------------------------------------------------------------
// Empfangene Zeichen werden in die Eingabgs-FIFO gespeichert und warten dort
// keine berlaufsicherung, da ja wir Master sind.
//

ISR(USART1_RXC_vect)
{
    unsigned char resh, resl;
 
 
 LED_LOCO_ON;   
 LED_LOCO_OFF;   


    if (UCSR1A & (1<< FE1))
	  { // Frame Error
        resh = UCSR1B;
        resl = UDR1;                // zumindest lesen!
	  } 
    else
      {
        if (UCSR1A & (1<< DOR1))
          { // DATA Overrun -> Fatal
            // !!! 
          }
        
        resh = UCSR1B;
        resh = (resh >> RXB81) & 0x01;
        resl = UDR1;
       	X_RxBuffer[X_rx_write_ptr] = (resh << 8) | resl;

        X_rx_write_ptr++;
        if (X_rx_write_ptr == X_RxBuffer_Size) X_rx_write_ptr=0;
        X_rx_fill++;          // no check for full, we just do an overrun!
      }
}

//----------------------------------------------------------------------------
// Ein Zeichen aus der Ausgabe-FIFO lesen und ausgeben
// Ist das Zeichen fertig ausgegeben, wird ein neuer SIG_UART_DATA-IRQ getriggert
// Ist das FIFO leer, deaktiviert die ISR ihren eigenen IRQ.

// We transmit 9 bits, but the 8th bit is allways zero.

#if (DEBUG == 3)
   unsigned  char X_tx_char;
#endif

ISR(USART1_UDRE_vect) 
  {

    if (X_tx_read_ptr != X_tx_write_ptr)
      {
        UCSR1A |= (1 << TXC1);              // writing a one clears any existing tx complete flag
        UCSR1B &= ~(1<<TXB81);               // clear bit 8
        
        UDR1 = X_TxBuffer[X_tx_read_ptr];
        #if (DEBUG == 3)
            X_tx_char = X_TxBuffer[X_tx_read_ptr];
        #endif
        X_tx_read_ptr++;
        if (X_tx_read_ptr == X_TxBuffer_Size) X_tx_read_ptr=0;
        X_tx_fill--;
      }
    else
        UCSR1B &= ~(1 << UDRIE1);           // disable further TxINT
  }



//=============================================================================
// Upstream Interface
//-----------------------------------------------------------------------------
// TX:
bool XP_tx_ready (void)
{
    if (X_tx_fill < (X_TxBuffer_Size-18))     // keep space for one complete message (16)
      {
        return(1);                        // true if enough room
      }
    else
      {
        return(0);
      }
}


// ret 1 if full
// This goes with fifo (up to 32), bit 8 is 0.

bool XP_send_byte (const unsigned char c)
{
    X_TxBuffer[X_tx_write_ptr] = c;

    X_tx_write_ptr++;
    if (X_tx_write_ptr == X_TxBuffer_Size) X_tx_write_ptr=0;

    unsigned char mysreg = SREG;
    cli();
    X_tx_fill++;
    SREG = mysreg;

    UCSR1B |= (1 << UDRIE1);   // enable TxINT

    if (X_tx_fill < (X_TxBuffer_Size-18))
      {
        return(1);
      }

    return(0);
}




// This goes without fifo (busy waiting), bit 8 is set.
// Bit 7 is masked and filled to get even parity

void XP_send_call_byte (const unsigned char c)
  {
    unsigned char my_c, temp;

    my_c = c & 0x7F;
    temp = my_c ^ (my_c << 4);
    temp = temp ^ (temp << 2);
    temp = temp ^ (temp << 1);
    my_c |= temp & 0x80;
   
    // alternative: my_c = c & 0x7F;
    // alternative: if (parity_even_bit(my_c)) my_c |= 0x80;

    // Wait for empty transmit buffer
    while ( !( UCSR1A & (1<<UDRE1)));

    set_XP_to_transmit();        // enable RS485 driver

    UCSR1B |= (1<<TXB81);        // set bit 8
    UDR1 = my_c;
  }

// ret 1 if all is sent
bool XP_is_all_sent (void)
{
    if (X_tx_fill == 0)
      {
        if (!(UCSR1A & (1 << UDRE1)))  return(0);    // UDR not empty
        if (!(UCSR1A & (1 << TXC1)))  return(0);    // TX Completed not set
        return(1);                        
      }
    else
      {
        return(0);
      }
}

void X_uart_puts (const char *s)
{
    do
      {
        XP_send_byte (*s);   // no check, only short strings
      }
    while (*s++);
}


//------------------------------------------------------------------------------
// RX:
bool XP_rx_ready (void)
{
    if (X_rx_read_ptr != X_rx_write_ptr)
      {
        return(1);     // there is something
      }
    else
      {
        return(0);  
      }
}

//-------------------------------------------------------------------
// rx_fifo_read gets one char from the input fifo
//
// there is no check whether a char is ready, this must be
// done before calling with a call to rx_fifo_ready();

unsigned int XP_rx_read (void)
  {
    unsigned int retval;

    retval = X_RxBuffer[X_rx_read_ptr];
    X_rx_read_ptr++;
    if (X_rx_read_ptr == X_RxBuffer_Size) X_rx_read_ptr=0;

    unsigned char mysreg = SREG;
    cli();
    X_rx_fill--;
    SREG = mysreg;
        
    return(retval);
  }


//-----------------------------------------------------------------------------------
// testroutinen zu Parsertest
// Ersatzroutine, um ein Zeichen in den Empfangsbuffer abzulegen
// Sendepuffer kann im Simulator mit einen Databreakpoint auf UDR berwacht werden.
void X_push_to_rx(unsigned int d)
  {
	X_RxBuffer[X_rx_write_ptr] = d;

    X_rx_write_ptr++;
    if (X_rx_write_ptr == X_RxBuffer_Size) X_rx_write_ptr=0;
    X_rx_fill++;
  }


